# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'xdcc_dl/gui/qt_designer/xdcc_downloader.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_XDCCDownloaderWindow(object):
    def setupUi(self, XDCCDownloaderWindow):
        XDCCDownloaderWindow.setObjectName("XDCCDownloaderWindow")
        XDCCDownloaderWindow.resize(928, 543)
        self.centralwidget = QtWidgets.QWidget(XDCCDownloaderWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.search_button = QtWidgets.QPushButton(self.centralwidget)
        self.search_button.setObjectName("search_button")
        self.gridLayout.addWidget(self.search_button, 7, 5, 1, 1)
        self.total_progress_bar = QtWidgets.QProgressBar(self.centralwidget)
        self.total_progress_bar.setProperty("value", 0)
        self.total_progress_bar.setObjectName("total_progress_bar")
        self.gridLayout.addWidget(self.total_progress_bar, 11, 9, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        self.gridLayout.addItem(spacerItem, 2, 0, 1, 1)
        self.message_edit = QtWidgets.QLineEdit(self.centralwidget)
        self.message_edit.setMaximumSize(QtCore.QSize(16777215, 50))
        self.message_edit.setObjectName("message_edit")
        self.gridLayout.addWidget(self.message_edit, 1, 0, 1, 6)
        spacerItem1 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 3, 8, 4, 1)
        self.line = QtWidgets.QFrame(self.centralwidget)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.gridLayout.addWidget(self.line, 9, 0, 1, 11)
        self.down_arrow_button = QtWidgets.QPushButton(self.centralwidget)
        self.down_arrow_button.setObjectName("down_arrow_button")
        self.gridLayout.addWidget(self.down_arrow_button, 7, 10, 1, 1)
        self.download_button = QtWidgets.QPushButton(self.centralwidget)
        self.download_button.setMaximumSize(QtCore.QSize(16777215, 20))
        self.download_button.setObjectName("download_button")
        self.gridLayout.addWidget(self.download_button, 11, 1, 1, 4)
        self.label_2 = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_2.sizePolicy().hasHeightForWidth())
        self.label_2.setSizePolicy(sizePolicy)
        self.label_2.setMaximumSize(QtCore.QSize(16777215, 20))
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 0, 6, 1, 4)
        self.single_progress_bar = QtWidgets.QProgressBar(self.centralwidget)
        self.single_progress_bar.setProperty("value", 0)
        self.single_progress_bar.setInvertedAppearance(False)
        self.single_progress_bar.setObjectName("single_progress_bar")
        self.gridLayout.addWidget(self.single_progress_bar, 10, 9, 1, 2)
        self.download_queue_list_widget = QtWidgets.QListWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(100)
        sizePolicy.setHeightForWidth(self.download_queue_list_widget.sizePolicy().hasHeightForWidth())
        self.download_queue_list_widget.setSizePolicy(sizePolicy)
        self.download_queue_list_widget.setMinimumSize(QtCore.QSize(350, 0))
        self.download_queue_list_widget.setSizeIncrement(QtCore.QSize(0, 100))
        self.download_queue_list_widget.setAlternatingRowColors(True)
        self.download_queue_list_widget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.download_queue_list_widget.setObjectName("download_queue_list_widget")
        self.gridLayout.addWidget(self.download_queue_list_widget, 3, 9, 4, 2)
        self.server_edit = QtWidgets.QLineEdit(self.centralwidget)
        self.server_edit.setMaximumSize(QtCore.QSize(16777215, 20))
        self.server_edit.setObjectName("server_edit")
        self.gridLayout.addWidget(self.server_edit, 1, 6, 1, 4)
        self.rigth_arrow_button = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.rigth_arrow_button.sizePolicy().hasHeightForWidth())
        self.rigth_arrow_button.setSizePolicy(sizePolicy)
        self.rigth_arrow_button.setMaximumSize(QtCore.QSize(40, 16777215))
        self.rigth_arrow_button.setObjectName("rigth_arrow_button")
        self.gridLayout.addWidget(self.rigth_arrow_button, 4, 7, 1, 1)
        self.search_engine_combo_box = QtWidgets.QComboBox(self.centralwidget)
        self.search_engine_combo_box.setObjectName("search_engine_combo_box")
        self.gridLayout.addWidget(self.search_engine_combo_box, 7, 4, 1, 1)
        self.label_3 = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_3.sizePolicy().hasHeightForWidth())
        self.label_3.setSizePolicy(sizePolicy)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 10, 0, 1, 1)
        self.label = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setMaximumSize(QtCore.QSize(16777215, 50))
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 6)
        self.destination_edit = QtWidgets.QLineEdit(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(100)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.destination_edit.sizePolicy().hasHeightForWidth())
        self.destination_edit.setSizePolicy(sizePolicy)
        self.destination_edit.setObjectName("destination_edit")
        self.gridLayout.addWidget(self.destination_edit, 10, 1, 1, 4)
        spacerItem2 = QtWidgets.QSpacerItem(20, 160, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem2, 6, 7, 1, 1)
        self.add_button = QtWidgets.QPushButton(self.centralwidget)
        self.add_button.setMaximumSize(QtCore.QSize(16777215, 50))
        self.add_button.setObjectName("add_button")
        self.gridLayout.addWidget(self.add_button, 1, 10, 1, 1)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem3, 3, 7, 1, 1)
        spacerItem4 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum)
        self.gridLayout.addItem(spacerItem4, 8, 4, 1, 1)
        self.search_term_edit = QtWidgets.QLineEdit(self.centralwidget)
        self.search_term_edit.setObjectName("search_term_edit")
        self.gridLayout.addWidget(self.search_term_edit, 7, 0, 1, 4)
        self.up_arrow_button = QtWidgets.QPushButton(self.centralwidget)
        self.up_arrow_button.setObjectName("up_arrow_button")
        self.gridLayout.addWidget(self.up_arrow_button, 7, 9, 1, 1)
        spacerItem5 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem5, 3, 6, 4, 1)
        self.left_arrow_button = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.left_arrow_button.sizePolicy().hasHeightForWidth())
        self.left_arrow_button.setSizePolicy(sizePolicy)
        self.left_arrow_button.setMaximumSize(QtCore.QSize(40, 16777215))
        self.left_arrow_button.setObjectName("left_arrow_button")
        self.gridLayout.addWidget(self.left_arrow_button, 5, 7, 1, 1)
        self.search_result_tree_widget = QtWidgets.QTreeWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(100)
        sizePolicy.setHeightForWidth(self.search_result_tree_widget.sizePolicy().hasHeightForWidth())
        self.search_result_tree_widget.setSizePolicy(sizePolicy)
        self.search_result_tree_widget.setMinimumSize(QtCore.QSize(455, 350))
        self.search_result_tree_widget.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.search_result_tree_widget.setSizeIncrement(QtCore.QSize(0, 100))
        self.search_result_tree_widget.setAlternatingRowColors(True)
        self.search_result_tree_widget.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.search_result_tree_widget.setObjectName("search_result_tree_widget")
        self.gridLayout.addWidget(self.search_result_tree_widget, 3, 0, 4, 6)
        self.label_4 = QtWidgets.QLabel(self.centralwidget)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 10, 6, 1, 3)
        self.label_5 = QtWidgets.QLabel(self.centralwidget)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 11, 6, 1, 3)
        XDCCDownloaderWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QtWidgets.QStatusBar(XDCCDownloaderWindow)
        self.statusbar.setObjectName("statusbar")
        XDCCDownloaderWindow.setStatusBar(self.statusbar)

        self.retranslateUi(XDCCDownloaderWindow)
        QtCore.QMetaObject.connectSlotsByName(XDCCDownloaderWindow)

    def retranslateUi(self, XDCCDownloaderWindow):
        _translate = QtCore.QCoreApplication.translate
        XDCCDownloaderWindow.setWindowTitle(_translate("XDCCDownloaderWindow", "XDCC Downloader"))
        self.search_button.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Start the search</p></body></html>"))
        self.search_button.setText(_translate("XDCCDownloaderWindow", "Search"))
        self.down_arrow_button.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Move down</p></body></html>"))
        self.down_arrow_button.setText(_translate("XDCCDownloaderWindow", "▼"))
        self.download_button.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Start the download</p></body></html>"))
        self.download_button.setText(_translate("XDCCDownloaderWindow", "Download"))
        self.label_2.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>The IRC Server to connect to</p></body></html>"))
        self.label_2.setText(_translate("XDCCDownloaderWindow", "Server"))
        self.download_queue_list_widget.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>This is the download queue.</p></body></html>"))
        self.server_edit.setText(_translate("XDCCDownloaderWindow", "irc.rizon.net"))
        self.rigth_arrow_button.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Add packs to queue</p></body></html>"))
        self.rigth_arrow_button.setText(_translate("XDCCDownloaderWindow", "►"))
        self.search_engine_combo_box.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>The Search Engine(s) to use</p></body></html>"))
        self.label_3.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Specify the target directory for the downloaded files</p></body></html>"))
        self.label_3.setText(_translate("XDCCDownloaderWindow", "Destination"))
        self.label.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>The XDCC Message copied from a packlist site</p></body></html>"))
        self.label.setText(_translate("XDCCDownloaderWindow", "Message"))
        self.add_button.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Adds the Pack to the Queue</p></body></html>"))
        self.add_button.setText(_translate("XDCCDownloaderWindow", "Add"))
        self.search_term_edit.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Enter the search term here</p></body></html>"))
        self.up_arrow_button.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Move up</p></body></html>"))
        self.up_arrow_button.setText(_translate("XDCCDownloaderWindow", "▲"))
        self.left_arrow_button.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>Remove items from queue</p></body></html>"))
        self.left_arrow_button.setText(_translate("XDCCDownloaderWindow", "◄"))
        self.search_result_tree_widget.setToolTip(_translate("XDCCDownloaderWindow", "<html><head/><body><p>The Search results are displayed here</p></body></html>"))
        self.search_result_tree_widget.setSortingEnabled(True)
        self.search_result_tree_widget.headerItem().setText(0, _translate("XDCCDownloaderWindow", "Index"))
        self.search_result_tree_widget.headerItem().setText(1, _translate("XDCCDownloaderWindow", "Bot"))
        self.search_result_tree_widget.headerItem().setText(2, _translate("XDCCDownloaderWindow", "Pack"))
        self.search_result_tree_widget.headerItem().setText(3, _translate("XDCCDownloaderWindow", "Size"))
        self.search_result_tree_widget.headerItem().setText(4, _translate("XDCCDownloaderWindow", "Filename"))
        self.label_4.setText(_translate("XDCCDownloaderWindow", "Single Progress"))
        self.label_5.setText(_translate("XDCCDownloaderWindow", "Total Progress"))




"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import random
import string


class User(object):
    """
    Models an IRC user
    """

    def __init__(self, username: str = "random") -> None:
        """
        Initializes the User

        :param username: the user's username. If left empty, or the string 'random' is passed, a random
                         username consisting only of ASCII characters will be generated as the username
                         An empty string will also result in a random username
        """
        if username == "random" or username == "":
            self.username = self.generate_random_username()
        else:
            self.username = username

    def get_name(self) -> str:
        """
        :return: The user's username
        """
        return self.username

    @staticmethod
    def generate_random_username(length: int = 10) -> str:
        """
        Generates a random username of given length

        :param length: The length of the username
        :return:       The random username
        """
        return "".join(random.choice(string.ascii_uppercase) for _ in range(length))



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import time


class Progress(object):
    """
    Class that keeps track of a download progress
    """

    def __init__(self, file_amount: int, callback: callable = None) -> None:
        """
        Creates a new Progress Object with a set amount of files

        :param file_amount: the files to download
        :param callback:    A callback method called whenever the progress' state is changed.
                            The callback will be given 8 parameters:
                                1: the single progress
                                2: the single progress size
                                3: the single progress completion percentage
                                4: the total progress completion percentage
                                5: the total progress
                                6: the total progress size
                                7: the current download speed in byte/s (Based on last 0.5 seconds)
                                8: the average download speed over the entire downloading process
        """
        self.start_time = time.time()
        self.speed_counter = time.time()

        self.previous_speed = 0
        self.speed_byte_counter = 0
        self.total_bytes = 0

        self.single_progress = 0
        self.single_total = 0
        self.total_progress = 0
        self.total_total = file_amount

        self.callback = callback

    def add_single_progress(self, new_bytes: int) -> int:
        """
        Adds a given amount of bytes to the single progress.

        :param new_bytes: the bytes to add
        :return:          the new single progress
        """
        self.single_progress += new_bytes
        self.total_bytes += new_bytes

        self.handle_callback()
        return self.single_progress

    def next_file(self) -> None:
        """
        Switches to the next file to download

        :return: None
        """
        if self.total_progress < self.total_total:
            self.total_progress += 1

            if self.total_progress != self.total_total:
                self.single_progress = 0
                self.single_total = 0

    def handle_callback(self) -> None:
        """
        Handles the callback method

        :return: None
        """
        if self.callback is not None:
            self.callback(self.single_progress, self.single_total, self.get_single_progress_percentage(),
                          self.total_progress, self.total_total, self.get_total_percentage(),
                          self.calculate_current_download_speed(), self.calculate_average_download_speed())

    def calculate_current_download_speed(self) -> int:
        """
        Calculates the current download speed based on the data downloaded every 0.5 seconds

        :return: The speed in byte/s
        """
        if time.time() - self.speed_counter < 0.5:
            return self.previous_speed
        else:
            bytes_downloaded = self.total_bytes - self.speed_byte_counter
            self.speed_byte_counter = self.total_bytes

            self.previous_speed = int(bytes_downloaded / (time.time() - self.speed_counter))
            self.speed_counter = time.time()

            return self.previous_speed

    def calculate_average_download_speed(self) -> int:
        """
        Calculates the average download speed during the entire run of the progress object

        :return: The average speed in byte/s
        """
        return int(self.total_bytes / (time.time() - self.start_time))

    def set_single_progress_total(self, total: int) -> None:
        """
        Sets the total amount of bytes of the single progress

        :param total: the total progress
        :return:      None
        """
        self.single_total = total

    def set_single_progress(self, progress: int) -> None:
        """
        Sets the single progress with the new file size

        :param progress: the new file size/progress of the download
        :return:         None
        """
        self.total_bytes += progress - self.single_progress
        self.single_progress = progress

        self.handle_callback()

    def get_single_progress_percentage(self) -> float:
        """
        :return: The percentage of completion of the single progress.
        """
        try:
            return (self.single_progress / self.single_total) * 100
        except ZeroDivisionError:
            return 0.0

    def get_total_percentage(self) -> float:
        """
        :return: The percentage of total progress
        """
        if self.total_progress == self.total_total and self.total_progress != 0:
            return 100.0

        try:
            total_percentage = self.total_progress / self.total_total
            single_percentage = (self.get_single_progress_percentage() / 100)
            single_proportional_percentage = single_percentage * (1 / self.total_total)

            total_percentage += single_proportional_percentage
            return total_percentage * 100
        except ZeroDivisionError:
            return 0.0

    def get_single_progress(self) -> int:
        """
        :return: The current single progress
        """
        return self.single_progress

    def get_single_progress_total(self) -> int:
        """
        :return: The single progress
        """
        return self.single_total



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import requests
from typing import List
from bs4 import BeautifulSoup
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.IrcServer import IrcServer


def find_ixirc_packs(search_phrase: str) -> List[XDCCPack]:
    """
    Searches for XDCC Packs matching the specified search string on ixirc.com

    :param search_phrase: The search phrase to search for
    :return:              The list of found XDCC Packs
    """

    if not search_phrase:
        return []

    # ixIRC.com replaces spaces with + symbols for search query URLs
    split_search_term = search_phrase.split(" ")
    prepared_search_term = split_search_term[0]
    i = 1
    while i < len(split_search_term):
        prepared_search_term += "+" + split_search_term[i]
        i += 1

    base_url = "https://ixirc.com/?q=" + prepared_search_term
    content = BeautifulSoup(requests.get(base_url).text, "html.parser")
    page_analysis = content.select("h3")
    page_contents = [content]  # This will be parsed for XDCC Packs later, along with the other pages

    # IxIRC only displays 30 results per page. Which is why we need to check how many pages there are in total.
    # If 'Over' is used in the h3 section of the page, it means that this is not the last
    # page with search results. It displays "Over X episodes" on all pages except the last, where
    # the exact amount of results is mentioned and the word 'Over' is omitted
    number_of_pages = 2 if "Over" in page_analysis[0].text else 1

    # Check for more pages and add their content to the page_contents list, only if more than one page exists
    analysing = (number_of_pages == 2)
    while analysing:

        # The new URL specifies a page number using &pn=
        url = "https://ixirc.com/?q=" + prepared_search_term + "&pn=" + str(number_of_pages - 1)
        content = BeautifulSoup(requests.get(url).text, "html.parser")
        page_contents.append(content)

        page_analysis = content.select("h3")
        if "Over" in page_analysis[0].text:
            # Found another page
            number_of_pages += 1
            continue
        else:
            # This was the last page
            analysing = False

    # Establish search results
    results = []
    for content in page_contents:
        results += get_page_results(content)

    results.sort(key=lambda x: x.packnumber)
    results.sort(key=lambda x: x.bot)
    return results


def get_page_results(page_content: BeautifulSoup) -> List[XDCCPack]:
    """
    This parses a single ixIRC page to find all search results from that page

    :param page_content: The Beautifulsoup-parsed content of the page
    :return:             A list of XDCC Packs on that page
    """
    results = []
    packs = page_content.select("td")

    # Initialize the pack variables
    file_name = ""
    bot = ""
    server = ""
    pack_number = 0
    size = ""

    column_count = 0      # Keeps track of which column the parser is currently working on

    # Counts how often the word "ago" was used, which is used to keep track of on which
    # pack we currently are. Each pack has two instances of 'ago' occurring.
    ago_count = 0

    aborted = False       # The process is aborted whenever an invalid pack is encountered
    next_element = False  # Flag that lets other parts of the loop know that we are moving on to the next pack

    # line_part is a x,y section of the rows and columns of the website. we go through it in the order
    # Left->Right, Top->Bottom
    for line_part in packs:

        if next_element and line_part.text == "":
            # Jumps to the next not-empty element if we are currently jumping to the next pack
            continue

        elif next_element and not line_part.text == "":
            # We reached a new pack, start parsing the new pack
            next_element = False

        elif not next_element and line_part.text == "":
            # Invalid pack element if a string == "" in the middle of the pack,
            # abort the pack and jump to next element
            aborted = True

        elif "ago" in line_part.text and column_count > 6:
            # Counts the number of times 'ago' is seen by the parser. The last two elements of a pack both end
            # with 'ago', which makes it ideal to use as a marker for when a single pack element ends.
            # This only starts counting once we got all relevant information from the pack itself (column_count > 6)
            # to avoid conflicts when the substring 'ago' is contained inside the file name
            ago_count += 1

        # This gets the information from the pack and stores them into variables
        # This gets skipped if it has been established that the pack is invalid
        if not aborted:
            if column_count == 0:
                file_name = line_part.text                         # File Name
            elif column_count == 1:
                server = "irc." + line_part.text.lower() + ".net"  # Server Address
            elif column_count == 2:
                pass  # Channel Information, not needed due to /whois IRC queries
            elif column_count == 3:
                bot = line_part.text                               # Bot Name
            elif column_count == 4:
                pack_number = int(line_part.text)                  # Pack Number
            elif column_count == 5:
                pass  # This is the 'gets' section, we don't need that
            elif column_count == 6:
                size = line_part.text                              # File Size

        # Resets state after a pack was successfully parsed, and adds xdcc pack to results
        if not aborted and ago_count == 2:
            ago_count = 0
            column_count = 0
            next_element = True

            # Generate XDCCPack and append it to the list
            result = XDCCPack(IrcServer(server), bot, pack_number)
            result.set_filename(file_name)
            result.set_size(size)
            results.append(result)

        # Resets state after invalid pack
        elif aborted and ago_count == 2:
            aborted = False
            ago_count = 0
            column_count = 0
            next_element = True

        if not next_element:
            # Only increment column_count in the middle of a pack, not when we jump to the next pack element
            column_count += 1

    return results



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import sys
import unittest
from xdcc_dl.main import main


class UnitTests(unittest.TestCase):

    def setUp(self):
        sys.argv = [sys.argv[0]]
        sys.argv.append("-v")
        sys.argv.append("0")
        sys.argv.append("-s")
        sys.argv.append("irc.namibsun.net")

    def tearDown(self):

        sys.argv = [sys.argv[0]]
        if os.path.isfile("1_test.txt"):
            os.remove("1_test.txt")
        if os.path.isfile("2_test.txt"):
            os.remove("2_test.txt")
        if os.path.isfile("3_test.txt"):
            os.remove("3_test.txt")

    def test_single_download(self):

        sys.argv.append("/msg xdcc_servbot xdcc send #1")
        main()

        self.check_content(1)

    def test_wrong_server(self):

        sys.argv.pop()
        sys.argv.pop()

        sys.argv.append("/msg xdcc_servbot xdcc send #1")
        self.assertFalse(os.path.isfile("1_test.txt"))

    def test_repeated_download(self):

        sys.argv.append("/msg xdcc_servbot xdcc send #2")

        with open("2_test.txt", 'w') as testtwo:
            testtwo.write("This is a Test File for XDCC File Transfers\n\n")
            testtwo.write("This is Pack 2")

        self.check_content(2)
        main()

        self.check_content(2)
        os.remove("2_test.txt")
        main()

        self.check_content(2)

    def test_resume(self):

        sys.argv.append("/msg xdcc_servbot xdcc send #3")

        with open("3_test.txt", 'w') as testthree:
            testthree.write("This is a Test File for XDCC File Transfers\n\nThis is Pack 3")

        self.check_content(3)

        with open("3_test.txt", 'rb') as testthree:
            binary = testthree.read()

        with open("3_test.txt", 'wb') as testthree:
            testthree.write(binary[0:int(len(binary) / 2)])

        main()
        self.check_content(3)

    def test_range_downloading(self):

        sys.argv.append("/msg xdcc_servbot xdcc send #1-3")
        main()

        self.check_content(1)
        self.check_content(2)
        self.check_content(3)

    def test_range_downloading_with_steps(self):

        sys.argv.append("/msg xdcc_servbot xdcc send #1-3,2")
        main()

        self.check_content(1)
        self.assertFalse(os.path.isfile("2_test.txt"))
        self.check_content(3)

    def check_content(self, number):

        self.assertTrue(os.path.isfile(str(number) + "_test.txt"))
        with open(str(number) + "_test.txt", 'r') as f:

            content = f.read()
            self.assertTrue("This is a Test File for XDCC File Transfers" in content)
            self.assertTrue("This is Pack " + str(number) in content)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import unittest
from xdcc_dl.logging.Logger import Logger
from xdcc_dl.logging.LoggingTypes import LoggingTypes


class UnitTests(unittest.TestCase):

    def setUp(self):
        self.logfile = os.path.join(os.getcwd(), "logging_log")
        self.messages = [("Invisible", LoggingTypes.INVISIBLE),
                         ("Default", LoggingTypes.DEFAULT),
                         ("WHOIS_SEND", LoggingTypes.WHOIS_SEND),
                         ("PRIVATE_NOTICE", LoggingTypes.PRIVATE_NOTICE),
                         ("MESSAGE_OF_THE_DAY", LoggingTypes.MESSAGE_OF_THE_DAY),
                         ("EVENT", LoggingTypes.EVENT),
                         ("CHANNEL_KICK", LoggingTypes.CHANNEL_KICK)]

    def tearDown(self):
        if os.path.isfile(self.logfile):
            os.remove(self.logfile)

    def test_verbosity(self, ignore_verbosity=False):

        for verbosity_level in range(0, 7):

            logger = Logger(verbosity_level, self.logfile, ignore_verbosity)

            for message in self.messages:
                logger.log(message[0], message[1])

            with open(self.logfile, 'r') as l:
                log = l.read().rstrip().lstrip()

            if ignore_verbosity:
                self.assertEqual(len(log.split("\n")), 7)
            else:
                self.assertEqual(len(log.split("\n")), verbosity_level + 1)

            os.remove(self.logfile)

    def test_ignore_verbosity(self):
        self.test_verbosity(True)

    def test_logger_with_existing_logfile(self):

        with open(self.logfile, 'w') as f:
            f.write("Testing Logger\n")

        logger = Logger(1, self.logfile)
        logger.log("Test")

        with open(self.logfile, 'r') as f:
            self.assertEqual(f.read(), "Testing Logger\nTest\n")



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import requests
from typing import List
from bs4 import BeautifulSoup
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.IrcServer import IrcServer


def find_namibsun_packs(search_phrase: str) -> List[XDCCPack]:
    """
    Searches for XDCC Packs matching the specified search string on irc.namibsun.net:8000

    :param search_phrase: The search phrase to search for
    :return:              The list of found XDCC Packs
    """

    soup = BeautifulSoup(requests.get("http://irc.namibsun.net:8000").text, "html.parser")
    content = soup.select(".content")

    all_packs = []
    for i, item in enumerate(content):
        if item.text == search_phrase:
            all_packs.append(XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", i + 1))
            all_packs[len(all_packs) - 1].set_filename(item.text)

    return all_packs



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
from xdcc_dl.metadata import PypiVariables
from setuptools import setup, find_packages


def readme():
    """
    Reads the readme file.

    :return: the readme file as a string
    """
    # noinspection PyBroadException
    try:
        # noinspection PyPackageRequirements
        import pypandoc
        with open('README.md') as f:
            # Convert markdown file to rst
            markdown = f.read()
            rst = pypandoc.convert(markdown, 'rst', format='md')
            return rst
    except:
        # If pandoc is not installed, just return the raw markdown text
        with open('README.md') as f:
            return f.read()


def find_scripts():
    """
    Returns a list of scripts in the bin directory

    :return: the list of scripts
    """
    try:
        scripts = []
        for file_name in os.listdir("bin"):
            if not file_name == "__init__.py" and os.path.isfile(os.path.join("bin", file_name)):
                scripts.append(os.path.join("bin", file_name))
        return scripts
    except OSError:
        return []

setup(name=PypiVariables.name,
      version=PypiVariables.version,
      description=PypiVariables.description,
      long_description=readme(),
      classifiers=PypiVariables.classifiers,
      url=PypiVariables.url,
      download_url=PypiVariables.download_url,
      author=PypiVariables.author,
      author_email=PypiVariables.author_email,
      license=PypiVariables.license,
      packages=find_packages(),
      install_requires=PypiVariables.install_requires,
      test_suite='nose.collector',
      tests_require=['nose'],
      scripts=find_scripts(),
      zip_safe=False)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import requests
from typing import List
from bs4 import BeautifulSoup
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.IrcServer import IrcServer


def find_intel_haruhichan_packs(search_phrase: str) -> List[XDCCPack]:
    """
    Searches for XDCC Packs matching the specified search string on intel.haruhichan.com

    :param search_phrase: The search phrase to search for
    :return:              The list of found XDCC Packs
    """

    if not search_phrase:
        return []

    # Generate the search string that can be inserted into intel.haruhichan.com's URL to conduct a search
    # Intel Haruhichan uses %20 to separate spaces in their search query URLs
    split_search_term = search_phrase.split(" ")
    prepared_search_term = split_search_term[0]
    i = 1
    while i < len(split_search_term):
        prepared_search_term += "%20" + split_search_term[i]
        i += 1

    url = "http://intel.haruhichan.com/?s=" + prepared_search_term
    content = BeautifulSoup(requests.get(url).text, "html.parser")
    packs = content.select("td")

    results = []

    bot = ""
    packnumber = ""
    size = ""

    for i, line in enumerate(packs):

        # Explanation how this works:
        # Every fifth 'td' element is the start of a new search result. They go in order:
        #       1. Bot Name
        #       2. Pack Number
        #       3. Requests (Not used)
        #       4. File Size
        #       5. File Name
        #
        # This means, that we check with modulo 5 at which td element we currently are, and increment i
        # after every loop.
        # If we are at the fifth element, we have all the information we need to generate an XDCCPack

        if i % 5 == 0:
            bot = line.text
        elif (i - 1) % 5 == 0:
            packnumber = int(line.text)
        elif (i - 2) % 5 == 0:
            pass    # Skip the 'requests' section
        elif (i - 3) % 5 == 0:
            size = line.text
        else:
            result = XDCCPack(IrcServer("irc.rizon.net"), bot, packnumber)
            result.set_size(size)
            result.set_filename(line.text)
            results.append(result)

    results.sort(key=lambda x: x.packnumber)
    results.sort(key=lambda x: x.bot)
    return results



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import unittest
from xdcc_dl.xdcc.layers.xdcc.MessageSender import MessageSender


class TestException(Exception):
    pass


class UnitTests(unittest.TestCase):

    def setUp(self):
        self.client = MessageSender

    def tearDown(self):
        self.client.quit()

    def test_logging(self):

        class Tester(MessageSender):
            def on_welcome(self, conn, event):

                self.on_currenttopic(conn, event)
                self.on_topicinfo(conn, event)
                self.on_quit(conn, event)
                self.on_part(conn, event)
                self.on_kick(conn, event)
                self.on_mode(conn, event)
                self.on_action(conn, event)
                self.on_nick(conn, event)

                raise TestException()

        self.client = Tester("irc.namibsun.net", "random")
        try:
            self.client.start()
            self.assertFalse(True)
        except TestException:
            self.assertTrue(True)

    def test_other_user_joined_channel(self):

        class Tester(MessageSender):
            def on_welcome(self, conn, event):

                event.source = "Other User"
                self.on_join(conn, event)

            def on_join(self, conn, event):
                super().on_join(conn, event)
                raise TestException()

        self.client = Tester("irc.namibsun.net", "random")
        try:
            self.client.start()
            self.assertFalse(True)
        except TestException:
            self.assertTrue(True)

    def test_joining_second_channel(self):

        class Tester(MessageSender):
            def on_welcome(self, conn, event):

                self.channel_joined = True
                event.source = self.user.get_name()
                self.on_join(conn, event)
                raise TestException()

        self.client = Tester("irc.namibsun.net", "random")
        try:
            self.client.start()
            self.assertFalse(True)
        except TestException:
            self.assertTrue(True)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import time
import struct
import irc.client
# noinspection PyPep8Naming
from xdcc_dl.logging.LoggingTypes import LoggingTypes as LOG
from xdcc_dl.xdcc.layers.xdcc.XDCCInitiator import XDCCInitiator


class AlreadyDownloaded(Exception):
    """
    Gets thrown if a file already exists with size >= download size
    """
    pass


# noinspection PyUnusedLocal
class DownloadHandler(XDCCInitiator):
    """
    Class that handles the download process
    Layer 5 of the XDCC Bot
    """

    def on_dccmsg(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Runs each time a new chunk of data is received while downloading

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        if self.already_downloaded:  # Send a single message to the server indicating that the transfer is done
            self.dcc_connection.send_bytes(struct.pack(b"!I", self.progress.get_single_progress_total()))

        else:

            data = event.arguments[0]
            data_length = len(data)

            self.file.write(data)
            self.progress.add_single_progress(data_length)

            progress_message = " Progress: %.2f" % self.progress.get_single_progress_percentage()
            progress_message += "% (" + str(self.progress.get_single_progress())
            progress_message += "/" + str(self.progress.get_single_progress_total()) + ")"

            self.logger.log(progress_message, LOG.DOWNLOAD_PROGRESS, carriage_return=True)

            # Send Acknowledge Message
            self.dcc_connection.send_bytes(struct.pack(b"!I", self.progress.get_single_progress()))
            # -> on_dccmsg if download not yet complete
            # -> on_dcc_disconnect if completed

    def on_dcc_disconnect(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        The DCC Connection was disconnected. Checks if download was completed. If not, try to resend Pack request

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        if self.file is not None:
            self.file.close()

            if os.path.getsize(self.current_pack.get_filepath()) < self.filesize:
                self.logger.log("Download Incomplete, Trying again.", LOG.DOWNLOAD_INCOMPLETE)
                self.connection.privmsg(self.current_pack.get_bot(), self.current_pack.get_request_message())
                return
            else:
                self.logger.log("\nDownload completed in %.2f seconds" % (time.time() - self.start_time), LOG.DEFAULT)

        if len(self.pack_queue) > 0:

            if not self.current_pack.get_server().get_address() == self.pack_queue[0].get_server().get_address():
                self.quit()  # -> on_disconnect

            else:
                self.pack_states[self.current_pack] = "OK"
                self.reset_connection_state()
                self.connected_to_server = True
                self.current_pack = self.pack_queue.pop(0)
                self.progress.next_file()
                self.connection.whois(self.current_pack.get_bot())

        else:
            self.quit()  # -> on_disconnect

    def on_disconnect(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Extends the disconnect method by adding a check for already downloaded files

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        if self.already_downloaded:
            raise AlreadyDownloaded()
        else:
            super().on_disconnect(connection, event)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
from typing import List, Dict

from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.Progress import Progress
from xdcc_dl.xdcc.layers.irc.BaseIrcClient import NetworkError
from xdcc_dl.xdcc.layers.irc.BotFinder import BotNotFoundException
from xdcc_dl.xdcc.layers.xdcc.DownloadHandler import DownloadHandler, AlreadyDownloaded
from xdcc_dl.xdcc.layers.xdcc.XDCCInitiator import IncorrectFileSentException, NoValidWhoisQueryException


class XDCCDownloader(DownloadHandler):
    """
    The XDCC Downloader that combines the capabilities of all XDCC Layers to offer a stable
    interface to download XDCC Packs
    """

    def download(self, packs: List[XDCCPack], progress: Progress = None) -> Dict[XDCCPack, str]:
        """
        Downloads all XDCC packs specified. Optionally shares state with other threads using a Progress object
        All packs need to connect to the same server

        :param packs:    The packs to download
        :param progress: Optional Progress object
        :return:         Dictionary of packs mapped to status codes:
                         "OK":              Download was successful
                         "BOTNOTFOUND":     Bot was not found
                         "CHANNELJOINFAIL": Channel join failed, most likely due to missing whois information
                         "NETWORKERROR":    Download failed due to network error
                         "INCORRECT":       Sent file was not the correct file
                         "EXISTED":         File already existed and was completely downloaded
                         "OTHERSERVER":     If a pack was found that is hosted on a different server
        """
        self.progress = progress if progress is not None else Progress(len(packs))
        self.pack_queue = packs
        self.pack_states = {}

        while len(self.pack_queue) > 0:

            self.current_pack = self.pack_queue.pop(0)

            if self.current_pack.get_server().get_address() != self.server.get_address():
                self.pack_states[self.current_pack] = "OTHERSERVER"
                continue

            status_code = "OK"

            try:
                self.start()
            except BotNotFoundException:
                status_code = "BOTNOTFOUND"
            except NoValidWhoisQueryException:
                status_code = "CHANNELJOINFAIL"
            except AlreadyDownloaded:
                status_code = "EXISTED"
            except IncorrectFileSentException:
                status_code = "INCORRECT"
            except NetworkError:
                status_code = "NETWORKERROR"

            self.pack_states[self.current_pack] = status_code
            self.reset_connection_state()
            self.progress.next_file()

        self.quit()

        return self.pack_states



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import unittest
from xdcc_dl.pack_searchers.PackSearcher import PackSearcher


class UnitTests(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_all_searchers(self):

        results = PackSearcher().search("Gin.txt")
        self.assertEqual(len(results), 4)

    def test_getting_searchers(self):

        procedures = PackSearcher.get_available_pack_searchers()
        searcher = PackSearcher()

        for procedure in procedures:
            self.assertTrue(PackSearcher.procedure_map[procedure] in searcher.procedures)

    def test_selected_searchers(self):

        searcher = PackSearcher(["nibl", "namibsun"])
        self.assertEqual(len(searcher.search("Gin.txt")), 1)

        for procedure in PackSearcher.get_available_pack_searchers():

            if procedure not in ["nibl", "namibsun"]:
                self.assertFalse(PackSearcher.procedure_map[procedure] in searcher.procedures)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
import unittest
from xdcc_dl.metadata import SentryLogger
from xdcc_dl.entities.XDCCPack import XDCCPack
from xdcc_dl.entities.Progress import Progress
from xdcc_dl.entities.IrcServer import IrcServer
from xdcc_dl.xdcc.XDCCDownloader import XDCCDownloader
from xdcc_dl.xdcc.layers.helpers.BotChannelMapper import BotChannelMapper


class UnitTests(unittest.TestCase):

    def setUp(self):
        self.downloader = XDCCDownloader("irc.namibsun.net", "random")

    def tearDown(self):
        if os.path.isfile("1_test.txt"):
            os.remove("1_test.txt")
        if os.path.isfile("2_test.txt"):
            os.remove("2_test.txt")
        if os.path.isfile("3_test.txt"):
            os.remove("3_test.txt")

    def test_download_multiple_packs(self):

        progress = Progress(2)

        packs = [XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2),
                 XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 3)]

        results = self.downloader.download(packs, progress)

        for pack in results:
            self.assertTrue(os.path.isfile(pack.get_filepath()))
            self.assertEqual(results[pack], "OK")

        self.assertEqual(progress.get_single_progress_percentage(), 100.0)
        self.assertEqual(progress.get_total_percentage(), 100.0)

    def test_bot_not_found(self):

        pack = XDCCPack(IrcServer("irc.namibsun.net"), "nosuchbot", 2)
        results = self.downloader.download([pack])

        self.assertFalse(os.path.isfile(pack.get_filepath()))
        self.assertEqual(results[pack], "BOTNOTFOUND")

    def test_file_existed(self):

        pack = XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2)

        self.downloader.download([pack])
        self.assertTrue(os.path.isfile(pack.get_filepath()))

        results = self.downloader.download([pack])
        self.assertTrue(os.path.isfile(pack.get_filepath()))
        self.assertEqual(results[pack], "EXISTED")

    def test_incorrect_file_sent(self):

        pack = XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2)
        pack.set_original_filename("something_else.txt")

        results = self.downloader.download([pack])
        self.assertFalse(os.path.isfile(pack.get_filepath()))
        self.assertEqual(results[pack], "INCORRECT")

    def test_network_error(self):

        pack = XDCCPack(IrcServer("gitlab.namibsun.net"), "xdcc_servbot", 2)
        results = XDCCDownloader("gitlab.namibsun.net", "random").download([pack])
        self.assertEqual(results[pack], "NETWORKERROR")

    def test_different_servers(self):

        packs_one = XDCCPack(IrcServer("irc.namibsun.net"), "xdcc_servbot", 2)
        pack_two = XDCCPack(IrcServer("irc.rizon.net"), "xdcc_servbot", 2)

        results = self.downloader.download([packs_one, pack_two])

        self.assertEqual(results[packs_one], "OK")
        self.assertEqual(results[pack_two], "OTHERSERVER")

    def test_invalid_whois_query(self):

        class DummySentry(object):
            # noinspection PyPep8Naming
            def captureMessage(self, string):
                pass

        BotChannelMapper.bot_channel_map = {}
        SentryLogger.sentry = DummySentry()

        pack = XDCCPack(IrcServer("irc.rizon.net"), "HelloKitty", 1)
        results = XDCCDownloader("irc.rizon.net", "random").download([pack])
        self.assertEqual(results[pack], "CHANNELJOINFAIL")



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import irc.client
from xdcc_dl.xdcc.layers.irc.BotFinder import BotFinder
# noinspection PyPep8Naming
from xdcc_dl.logging.LoggingTypes import LoggingTypes as LOG


# noinspection PyUnusedLocal
class MessageSender(BotFinder):
    """
    Class that handles initiating the XDCC Connection by sending the appropriate message to the bot
    Layer 3 of the XDCC Bot
    """

    def on_join(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Method called when a channel was joined. The first channel that was joined triggers sending a message
        to the bot to initiate the XDCC handshake

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        if event.source.startswith(self.user.get_name()):   # Check if we were the ones joining
            if not self.user.get_name() == event.target and self.channel_join_required:
                # Check if we didn't ask a channelless bot
                self.logger.log("Joined Channel " + event.target, LOG.CHANNEL_JOIN_SUCCESS)

            if not self.channel_joined:  # Only send the XDCC message when the first channel was joined
                self.channel_joined = True  # Let other on_joins know that message was already sent

                log_message = "Sending XDCC Request to " + self.current_pack.get_bot()
                log_message += " for pack " + str(self.current_pack.get_packnumber())
                self.logger.log(log_message, LOG.MESSAGE_SEND)

                connection.privmsg(self.current_pack.get_bot(), self.current_pack.get_request_message())
                # -> on_ctcp

    def on_namreply(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Event when a channel sends the existing users on the channel after joining

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        users = event.arguments[2].split(" ")
        message = "Channel Users:\n"
        for user in users:
            message += user + ";"
        self.logger.log(message, LOG.CHANNEL_USERS)

    def on_endofnames(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Marks the End of a series of namreplys

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log("End Of joined users list", LOG.CHANNEL_USERS)

    def on_currenttopic(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Receives the current topic of the newly joined channel

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log("Topic " + str(event.arguments), LOG.CHANNEL_TOPIC)

    def on_topicinfo(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Receives the current topic information of the newly joined Channel

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log("Topic Info: " + str(event.arguments), LOG.CHANNEL_TOPIC)

    def on_quit(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Logs that a user left a channel

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.arguments[0] + " quit.", LOG.CHANNEL_QUIT)

    def on_part(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Logs that a user left a channel using part

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.source + " left.", LOG.CHANNEL_PART)

    def on_kick(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Logs that a user was kicked from the channel

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.arguments[0] + " was kicked.", LOG.CHANNEL_KICK)

    def on_mode(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Called whenever the mode of a user changes

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.arguments[0] + " mode changed", LOG.CHANNEL_MODE_CHANGE)

    def on_action(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Called whenever an action on the channel occurs

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.arguments[0] + " mode changed", LOG.CHANNEL_ACTION)

    def on_nick(self, connection: irc.client.ServerConnection, event: irc.client.Event) -> None:
        """
        Called whenever a user changes his nick

        :param connection: the IRC Connection
        :param event:      the IRC Event
        :return:           None
        """
        self.logger.log(event.source + " nick changed", LOG.CHANNEL_NICK_CHANGED)



"""
LICENSE:
Copyright 2016 Hermann Krumrey

This file is part of xdcc_dl.

    xdcc_dl is a program that allows downloading files via the XDCC
    protocol via file serving bots on IRC networks.

    xdcc_dl is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    xdcc_dl is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with xdcc_dl.  If not, see <http://www.gnu.org/licenses/>.
LICENSE
"""

# imports
import os
from typing import Dict
from xdcc_dl.logging.LoggingTypes import LoggingTypes


class Logger(object):
    """
    Logger for IRC/XDCC related events with variable verbosity levels
    """

    def __init__(self, verbosity_level: int, logfile: str = None, ignore_verbosity_in_logfile: bool = True) -> None:
        """
        Initializes a new Logger object. The verbosity of the logger is defined here.
        Optionally, a logfile can be used to log to a file in addition to the console.
        If a logfile is specified, everything, regardless of verbosity level, will be logged
        to the file by default, but may be changed by toggling the ignore_verbosity_in_logfile parameter

        :param verbosity_level:             the verbosity level to be used. for more information, see
                                            xdcc_dl.logging.LoggingTypes
        :param logfile:                     The path to a logfile. Specifying this is optional
        :param ignore_verbosity_in_logfile: Determines if the logfile uses the same verbosity level settings
                                            as the console logging. By default, the verbosity setting is ignored
        """
        self.verbosity_level = verbosity_level
        self.logfile = logfile
        self.ignore_logfile_verbosity = ignore_verbosity_in_logfile

        if self.logfile is not None:
            if not os.path.isfile(self.logfile):
                open(self.logfile, 'w').close()

    def log(self, message: str, logging_type: Dict[str, str or int]=None, carriage_return: bool = False) -> None:
        """
        Logs a message to the console and optionally to a logfile

        :param message:         the message to log
        :param logging_type:    the type of message to log
        :param carriage_return: If set to true, appends a carriage return to the log message
        :return:                None
        """
        if logging_type is None:
            logging_type = LoggingTypes.DEFAULT

        priority = logging_type['priority']

        if self.logfile is not None:
            if self.ignore_logfile_verbosity or self.verbosity_level >= priority:
                with open(self.logfile, 'a') as log:
                    log.write(message + "\n")

        if self.verbosity_level >= priority:

            fg_color = logging_type["fg_color"]
            bg_color = logging_type["bg_color"]

            end = "\n" if not carriage_return else "\r"

            print(fg_color + bg_color + message + '\033[0m' + '\033[0m', end=end)
